import torch.utils.data as data
import torch
from torchvision import transforms
import torch.nn as nn
from torch.optim import Adam
from pathlib import Path
import matplotlib
import matplotlib.pyplot as plt
import torchvision.transforms.functional as TF
from open3d import *

import torchvision
from loader import VisdomLoader
from models import get_model
from transform import binarize, image_to_tensor
from visualizer import visualize_tensor, visualize_mask, denormalize
from pcgen import point_cloud_generator, visualize_pcd
import numpy as np
import time

input_transform = transforms.Compose([
    transforms.CenterCrop((768, 1024)),
    transforms.Resize((480, 640)),
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.2092, 0.2172, 0.2155], std=[0.1872, 0.1733, 0.1625]),
])

target_transform = transforms.Compose([
    transforms.CenterCrop((768, 1024)),
    transforms.Resize((480, 640)),
    binarize()
    #transforms.ToTensor()
])

depth_transform = transforms.Compose([
    transforms.CenterCrop((768, 1024)),
    transforms.Resize((480, 640)),
    transforms.ToTensor()
])


def load_zdf(fname):
    f = Dataset(fname)
    pcd = f['data']['pointcloud'][:,:,:]
    rgb = f['data']['rgba_image'][:,:,:3]
    contrast = f['data']['contrast'][:,:]
    depth = pcd[:, :, 2]
    f.close()
    return np.dstack([pcd,contrast,rgb]), rgb, depth


def load_model_state(root, fold):
    root = Path(root)
    model_path = root / 'model_{fold}.pt'.format(fold=fold)

    if model_path.exists():
        state = torch.load(str(model_path))
        model.load_state_dict(state['model'])
        print('Model state loaded!')
    else:
        print('Model does not have any model parameters...\n')


def crop_center(img,cropx,cropy):
    y,x = img.shape
    startx = x//2-(cropx//2)
    starty = y//2-(cropy//2)
    return img[starty:starty+cropy,startx:startx+cropx]


if __name__ == '__main__':
    #MODEL PARAMETERS
    num_classes = 2 #26
    batch_size = 8 #8
    learning_rate = 1e-4
    train_val_ratio = 0.8

    #LOAD DATA
    data_dir = './datasets/wisdom/wisdom-real/high-res'
    data_ds = VisdomLoader(data_dir, input_transform=input_transform,
                           target_transform=target_transform,
                           depth_transform=depth_transform)

    train_size = int(train_val_ratio * len(data_ds))
    val_size = len(data_ds) - train_size
    train_dataset, val_dataset = data.random_split(data_ds, [train_size, val_size])

    train_loader = data.DataLoader(train_dataset, batch_size=batch_size, shuffle=False, num_workers=0)
    val_loader = data.DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=0)

    model = get_model(num_classes=num_classes)
    optimizer = Adam(model.parameters(), lr=learning_rate)
    criterion = nn.CrossEntropyLoss()

    root = './history/binary'
    fold = 'binary-None_full_img2'
    load_model_state(root, fold)


    ### LOAD IMAGE TAKEN FROM ZIVID
    from netCDF4 import Dataset

    pcd, rgb, depth2 = load_zdf('C:/Users/cpascheh/Desktop/workspace/rgb/zivid_data/06052019/pcd1.zdf')

    ### IMAGE TRANSFORMATION
    image2 = TF.to_pil_image(rgb)

    trans = transforms.Compose([transforms.CenterCrop((1184, 1920)),
                        #transforms.Resize((480, 640)),
                        transforms.ToTensor()
                        ])
    rgb = trans(image2)
    #plt.imshow(rgb.numpy().transpose(1, 2, 0))
    #plt.show()

    ### MODEL PREDICTION
    model.eval()

    image, depth, gt_mask = val_loader.dataset[0]
    #visualize_mask(mask, title='Ground truth')
    #visualize_tensor(image, depth)


    #must unsqueeze for batch dim
    input = image.unsqueeze(0)
    input2 = rgb.unsqueeze(0).float()
    start = time.time()
    output = model(input)
    end = time.time()
    prediction = torch.argmax(output, dim=1)
    print('Time: ', end - start)
    visualize_mask(prediction, title='Network prediction')

    #image = denormalize(image)
    img = image.numpy().transpose(1,2,0)


    #plt.imshow(image2)
    #plt.show()

    visualize_tensor(image, depth, prediction, True)
    #plt.imshow(depth.squeeze())
    #plt.show()
    #plt.imshow(img)
    #plt.show()

    i = 1
    #torchvision.utils.save_image(gt_mask[0, :, :], './results/images/gt_mask_' + str(i) +'.png')
    #torchvision.utils.save_image(image,  './results/images/input_color_' + str(i) + '.png')
    #torchvision.utils.save_image(depth[0, :, :].squeeze(),  './results/images/input_depth_' + str(i) + '.png')
    #torchvision.utils.save_image(prediction[0, :, :],  './results/images/pred_mask_' + str(i) + '.png')


"""
    ### GENERATE 3D POINT CLOUD FROM OUTPUT

    cam_mat = PinholeCameraIntrinsic(1032, 772, 1105.0, 1105.0, 511.5, 384.0)

    image = denormalize(image)
    image = image.numpy().transpose(1, 2, 0)
    color = np.array(image)
    depth = depth.numpy().transpose(1, 2, 0)
    depth = Image((depth))
    color = color.astype(np.uint8)
    color = Image(color)

    #seg_image = mask_pc_overlay(color, mask)

    pcd = point_cloud_generator(color, depth, cam_mat)
    #visualize_pcd(pcd)
"""