import os
import os.path
import torch.utils.data as data
from PIL import Image
from transform import augmentation

EXTENSIONS = ['.jpg', '.png']


def is_image(filename):
    return any(filename.endswith(ext) for ext in EXTENSIONS)


def image_basename(filename):
    return os.path.basename(os.path.splitext(filename)[0])


def image_path(root, basename, extension):
    return os.path.join(root, f'{basename}{extension}')


class VisdomLoader(data.Dataset):

    def __init__(self, root, input_transform=None, target_transform=None, depth_transform=None):
        super().__init__()

        self.images_root = os.path.join(root, 'color_ims')
        self.masks_root = os.path.join(root, 'modal_segmasks')
        self.depth_root = os.path.join(root, 'depth_ims')


        self.filenames = [image_basename(f)
                          for f in os.listdir(self.images_root) if is_image(f)]

        self.input_transform = input_transform
        self.target_transform = target_transform
        self.depth_transform = depth_transform

    def __getitem__(self, index):
        filename = self.filenames[index]

        with open(image_path(self.images_root, filename, '.png'), 'rb') as f:
            image = Image.open(f).convert('RGB')
        with open(image_path(self.masks_root, filename, '.png'), 'rb') as f:
            mask = Image.open(f).convert('L')
        with open(image_path(self.depth_root, filename, '.png'), 'rb') as f:
            depth = Image.open(f).convert('L')

        image, mask, depth = augmentation(image, mask, depth)

        if self.input_transform is not None:
            image = self.input_transform(image)
        if self.target_transform is not None:
            mask = self.target_transform(mask)
        if self.depth_transform is not None:
            depth = self.depth_transform(depth)

        return image, depth, mask

    def __len__(self):
        return len(self.filenames)

