import numpy as np
import cv2
from open3d import *
import matplotlib.pyplot as plt
from load_zivid_file import load_zdf
from visualizer import apply_mask, display_instances, random_colors
import pickle
import pptk
import copy
import time
import os
from pathlib import Path




########## POSE ESTIMATION ################
# code adapted from open3d

def draw_registration_result(source, target, rotation=False):
    def rotate_view(vis):
        ctr = vis.get_view_control()
        ctr.rotate(5.0, 0.0)
        return False

    source_temp = copy.deepcopy(source)
    target_temp = copy.deepcopy(target)
    target_temp.paint_uniform_color([0.706, 1, 0])
    source_temp.paint_uniform_color([1, 0, 0])

    if rotation is True:
        draw_geometries_with_animation_callback([source_temp, target_temp], rotate_view)
    else:
        draw_geometries([source_temp, target_temp])


def preprocess_point_cloud(pcd, voxel_size):
    pcd_down = voxel_down_sample(pcd, voxel_size)

    radius_normal = voxel_size * 2
    estimate_normals(pcd_down, KDTreeSearchParamHybrid(
            radius = radius_normal, max_nn = 30))

    radius_feature = voxel_size * 5
    pcd_fpfh = compute_fpfh_feature(pcd_down,
            KDTreeSearchParamHybrid(radius = radius_feature, max_nn = 100))
    return pcd_down, pcd_fpfh


def prepare_dataset(source, target, voxel_size, trans_init):
    target.transform(trans_init)
    draw_registration_result(source, target)

    source_down, source_fpfh = preprocess_point_cloud(source, voxel_size)
    target_down, target_fpfh = preprocess_point_cloud(target, voxel_size)

    return source, target, source_down, target_down, source_fpfh, target_fpfh


def execute_fast_global_registration(source_down, target_down, source_fpfh, target_fpfh, voxel_size):
    distance_threshold = 1
    result = registration_fast_based_on_feature_matching(
            source_down, target_down, source_fpfh, target_fpfh,
            FastGlobalRegistrationOption(maximum_correspondence_distance = distance_threshold))
    return result


#Self implemented functions

def draw_centroid(image, cX, cY, class_ids, class_names, i):
    """
    Draws the centroid (cX,cY) and labels the object for a given image
    :param image: image to be drawn on
    :param cX: x-coordinates (pixels)
    :param cY: y-coordinates (pixels)
    :param class_ids: id number for given object
    :param class_names: list of all possible class names
    :return: image with centroid
    """
    # Display right category on centeroid.
    class_id = class_ids
    label = class_names[class_id]
    caption = "{}:{}".format(i, label)

    cv2.circle(image, (cX, cY), 5, (255, 0, 0), -1)
    cv2.putText(image, caption, (cX - 25, cY - 25), cv2.FONT_HERSHEY_SIMPLEX, 1.0, (255, 255, 255), 2)

    return image


def remove_zeros(pcd):
    """
    Remove NaNs from a point cloud
    :param pcd: input point cloud
    :return: same pointcloud without nans. (changes dim of point cloud)
    """
    x = np.extract(pcd[:, :, 2] != 0, pcd[:, :, 0])
    y = np.extract(pcd[:, :, 2] != 0, pcd[:, :, 1])
    z = np.extract(pcd[:, :, 2] != 0, pcd[:, :, 2])
    r = np.extract(pcd[:, :, 2] != 0, pcd[:, :, 3])
    g = np.extract(pcd[:, :, 2] != 0, pcd[:, :, 4])
    b = np.extract(pcd[:, :, 2] != 0, pcd[:, :, 5])

    return np.dstack([x,y,z,r,g,b])


def segment_pcd(pcd, mask):
    """
    Extracts points from a point cloud based on a segmentation mask
    :param pcd: input pointcloud
    :param mask: input mask
    :return: segmented point cloud
    """
    x = np.extract(mask, pcd[:, :, 0])
    y = np.extract(mask, pcd[:, :, 1])
    z = np.extract(mask, pcd[:, :, 2])
    r = np.extract(mask, pcd[:, :, 3])
    g = np.extract(mask, pcd[:, :, 4])
    b = np.extract(mask, pcd[:, :, 5])

    return np.dstack([x, y, z, r, g, b])


def calc_centroid_3d(pcd, mask):
    """
    Calculates the Center of Mass for a point cloud based on a mask
    :param pcd: input point cloud
    :param mask: mask weight
    :return: (x,y,z)-coordinates of CoM
    """
    cx = np.average(pcd[:,:,0], weights=mask)
    cy = np.average(pcd[:,:,1], weights=mask)
    cz = np.average(pcd[:,:,2], weights=mask)

    return cx, cy, cz


def get_pixel_positons(image):
    """
    Get pixel posistion given the pixel value
    :param image: input image
    :return: ndarray of pixel coordinates
    """
    yx = np.argwhere(image == 1)
    xy = np.flip(yx, axis=1)

    return xy


if __name__ == '__main__':

    ### Load network output and zivid data###
    img_nr = 1
    r = pickle.load(open('C:/Users/cpascheh/Desktop/workspace/rgb/zivid_data/results/03052019/results_image' + str(img_nr) + '.plk',
                         'rb'))
    pcd, rgb, depth = load_zdf('C:/Users/cpascheh/Desktop/workspace/rgb/zivid_data/03052019/pcd' + str(img_nr) + '.zdf')

    #Preprocessing point cloud
    pcd[np.isnan(pcd[:, :, 2])] = 0
    pcd = pcd[:, :, [0, 1, 2, 4, 5, 6]]

    class_names = ['BG', 'person', 'bicycle', 'car', 'motorcycle', 'airplane',
                   'bus', 'train', 'truck', 'boat', 'traffic light',
                   'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird',
                   'cat', 'dog', 'horse', 'sheep', 'cow', 'elephant', 'bear',
                   'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie',
                   'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball',
                   'kite', 'baseball bat', 'baseball glove', 'skateboard',
                   'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup',
                   'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple',
                   'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza',
                   'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed',
                   'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote',
                   'keyboard', 'cell phone', 'microwave', 'oven', 'toaster',
                   'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors',
                   'teddy bear', 'hair drier', 'toothbrush']


    masks = r['masks'].astype(np.uint8)
    colors = random_colors(masks.shape[2])

    seg_masks = np.zeros((masks.shape[0], masks.shape[1]))
    masked_image = rgb.astype(np.uint8).copy()
    for i in range(masks.shape[2]):
        mask = r['masks'][:, :, i]
        seg_masks = np.where(mask == 1, i + 1, seg_masks)
        masked_image = apply_mask(masked_image, mask, colors[i])

    #plt.imshow(masked_image)
    #plt.show()
    #display_instances(img_nr, rgb, r['rois'], r['masks'], r['class_ids'], class_names, r['scores'])

    pts = pcd.reshape(-1, 6)
    color = pts[:, 3:]
    seg_color = masked_image.reshape(-1, 3)
    #v = pptk.viewer(pts[:, :3])
    #v.attributes(color / 255., seg_color / 255.)

    for i in range(masks.shape[2]):
        mask = masks[:, :, i]
        seg_pcd = segment_pcd(pcd, mask)
        seg_pcd = remove_zeros(seg_pcd)

        #v = pptk.viewer(seg_pcd[:, :, :3])
        #v.attributes(seg_pcd[:, :, 3:].squeeze() / 255.)



    ### SYNTHETIC DATA ###
    time_list = []

    for instance in os.listdir('C:/Users/cpascheh/Desktop/workspace/rgb/sim_data/network_output/'):
        instance = '215.plk'
        print('INSTANCE: ', instance[:3])
        r = pickle.load(open('C:/Users/cpascheh/Desktop/workspace/rgb/sim_data/network_output/' + str(instance), 'rb'))
        pcd = np.load('C:/Users/cpascheh/Desktop/workspace/rgb/sim_data/pcds/pcd' + str(instance[:3]) + '.npy')
        rgb = pcd[:,:,3:]

        masks = r['masks'].astype(np.uint8)
        colors = random_colors(masks.shape[2])

        seg_masks = np.zeros((masks.shape[0], masks.shape[1]))
        masked_image = rgb.astype(np.uint8).copy()
        for i in range(masks.shape[2]):
            mask = r['masks'][:, :, i]
            seg_masks = np.where(mask == 1, i+1, seg_masks)
            masked_image = apply_mask(masked_image, mask, colors[i])

        #plt.imshow(masked_image)
        #plt.show()
        #display_instances(rgb, r['rois'], r['masks'], r['class_ids'], class_names, r['scores'])

        fig, ax = plt.subplots(1, figsize=(16, 14))
        plt.imshow(masked_image)
        ax.set_title('Segmented masks overlay', fontsize=22)
        plt.xlabel('pixels', fontsize=22)
        plt.ylabel('pixels', fontsize=22)
        plt.show()
        # plt.savefig('C:/Users/cpascheh/Desktop/master-plots/iou_gt_mask1.png')


        pts = pcd.reshape(-1, 6)
        color = pts[:, 3:]
        seg_color = masked_image.reshape(-1, 3)


        #draw_geometries([test1])
        #v = pptk.viewer(pcd[:, :, :3])
        #v.attributes(color / 255., seg_color / 255.)


        data = []
        for i in range(masks.shape[2]):
            mask = masks[:, :, i]
            seg_pcd = segment_pcd(pcd, mask)
            seg_pcd = remove_zeros(seg_pcd)

            # extract only valid values from the mask...
            xyz = seg_pcd[:, :, :3].squeeze()
            print('[3D] Object nr ', i)
            mean = xyz.mean(axis=0)
            print('MEAN: ', mean)

            #v = pptk.viewer(seg_pcd[:, :, :3])
            #v.attributes(seg_pcd[:, :, 3:].squeeze() / 255.)

            #test = PointCloud()
            #seg_pcd = seg_pcd.squeeze()
            #test.points = Vector3dVector(seg_pcd[:, :3])
            #test.colors = Vector3dVector(seg_pcd[:, 3:] / 255)
           #draw_geometries([test])

        #Statistics:
        #print('CENTROID comparison: ')
        #print('2D: ', pcd[cy, cx, :3], '  3D: ', mean)



        #------------------------------------------------
        #---------       Pose estimation        ---------
        #------------------------------------------------
            if r['class_ids'][i] == 77:
                target = read_point_cloud("./zivid_data/pose_est/gt_pcds/gt_scissor.ply")
            else:
                target = read_point_cloud("./zivid_data/pose_est/gt_pcds/gt_banana.ply")



            source = PointCloud()
            seg_pcd = seg_pcd.squeeze()
            source.points = Vector3dVector(seg_pcd[:, :3])
            source.colors = Vector3dVector(seg_pcd[:, 3:] / 255)


            trans_init = np.asarray([[-1.0, 0.0, 0.0, mean[0]],
                                    [0.0, 1.0, 0.0, mean[1]],
                                    [0.0, 0.0, -1.0, mean[2]],
                                    [0.0, 0.0, 0.0, 1.0]])


            voxel_size = 4
            source, target, source_down, target_down, source_fpfh, target_fpfh = prepare_dataset(source, target, voxel_size, trans_init)

            start = time.time()
            result_fast = execute_fast_global_registration(source_down, target_down,
                                                           source_fpfh, target_fpfh, voxel_size)

            timer = (time.time() - start)
            print("Fast global registration took %.3f sec.\n" % (time.time() - start))
            source_down.transform(result_fast.transformation)


            trans = np.dot(result_fast.transformation, trans_init)
            print(np.round(trans,3))
            print(result_fast)
            data.append(trans)

            time_list.append(timer)

            #draw_registration_result(source_down, target_down, rotation=False)

        save_path = './sim_data/pose_est/' + str(instance[:3])
        if not Path(save_path).exists():
            os.mkdir(save_path)
        #obj_name = class_names[r['class_ids'][i]]
        #np.save(save_path + '/' +'pose_est.npy', data)

        break
