from __future__ import print_function
from __future__ import division
import torch
from pathlib import Path
import copy
import tqdm
import numpy as np
import json
from datetime import datetime
from visualizer import save_images


def compute_batch_iou(outputs, targets):
    """
    Calculates the IoU for a batch of masks
    :param outputs:
    :param targets:
    :return:
    """
    intersect = (outputs * targets).sum(2).sum(1)
    union = (outputs + targets).sum(2).sum(1)
    iou = (intersect + 0.00001) / (union - intersect + 0.00001)

    return iou.numpy()


def compute_iou(mask1, mask2):
    """
    Calculates IoU between two binary masks
    :param mask1:
    :param mask2:
    :return:
    """
    if mask1.shape[-1] == 0 or mask2.shape[-1] == 0:
        return 0

    # flattern mask and compute area
    mask1 = np.reshape(mask1 > .5, -1).astype(np.float32)
    mask2 = np.reshape(mask2 > .5, -1).astype(np.float32)
    area1 = np.sum(mask1)
    area2 = np.sum(mask2)

    #avoid dividing with 0
    epsilon = 1e-15

    # intersection and union
    intersection = np.dot(mask1.T, mask2)
    union = area1 + area2 - intersection
    iou = (intersection + epsilon) / (union + epsilon)

    return iou


def binary_validation(model, criterion, valid_loader, device):
    with torch.no_grad():
        model.eval()
        losses = []
        iou_loss = []

        for inputs, depths, targets in valid_loader:
            inputs = inputs.to(device)
            targets = targets.to(device)

            outputs = model(inputs)

            out_mask = torch.argmax(outputs, dim=1).float()

            #plt.imshow(out_mask[0, :, :])
            #plt.title("Out mask")
            #plt.show()

            #plt.imshow(targets[0, :, :, :].squeeze())
            #plt.title("Gt mask")
            #plt.show()
            #visualize_tensor(i, epoch, inputs[0, :, :, :], depths[0, :, :], out_mask[0, :, :], True)


            loss = criterion(outputs, targets.squeeze().long())
            losses.append(loss.item())

            iou = compute_batch_iou(out_mask, targets.squeeze().float())
            print('[iou] ', iou)
            iou_loss.append(iou)


        valid_loss = np.mean(losses)

        valid_iou = np.mean(iou_loss).astype(np.float64)

        print('Valid iou: {:.5f}, loss: {:.5f}'.format(valid_iou, valid_loss))
        return valid_loss, valid_iou



def write_event(log, batch, **data):
    data['batch'] = batch
    data['date and time'] = datetime.now().isoformat()
    log.write(json.dumps(data, sort_keys=True))
    log.write('\n')
    log.flush()


def train_model(model, train_loader, valid_loader, criterion, optimizer, device, lr, batch_size,
                num_classes, root, num_epochs=10, fold=None):

    report_each = 2
    train_loss_history = []
    train_losses = []
    valid_losses = []
    best_loss = np.Inf

    #Checks if a trained model already exists
    root = Path(root)
    model_path = root / 'model_{fold}.pt'.format(fold=fold)

    if model_path.exists():
        print("[TRAINING]: A model exists \n")
        state = torch.load(str(model_path))
        epoch = state['epoch']+1
        batch = state['batch']
        model.load_state_dict(state['model'])
        print('[TRAINING]: Restored model, epoch {}, step {:,}\n'.format(epoch, batch))
    else:
        epoch = 0
        batch = 0


    save = lambda ep: torch.save({
        'model': model.state_dict(),
        'epoch': ep,
        'batch': batch,
    }, str(model_path))

    log = root.joinpath('train_{fold}.log'.format(fold=fold)).open('at', encoding='utf8')

    running_loss = 0.0
    for epoch in range(epoch, num_epochs):
        model.train()

        tq = tqdm.tqdm(total=(len(train_loader) * batch_size))
        tq.set_description('Epoch {}, lr {}'.format(epoch, lr))
        print('\n')

        epoch_loss = []
        for i, (inputs, depths, labels) in enumerate(train_loader):
            inputs = inputs.to(device)

            with torch.no_grad():
                labels = labels.to(device)


            outputs = model(inputs)

            batch_loss = criterion(outputs, labels.squeeze().long())
            print('[TRAINING LOSS]: ', batch_loss.item(), '\n')
            optimizer.zero_grad()
            batch_loss.backward()
            optimizer.step()
            tq.update(batch_size)

            # statistics
            running_loss += batch_loss.item()
            epoch_loss.append(batch_loss.item())
            train_losses.append(batch_loss.item())

            tq.set_postfix(batch_loss='{:.5f}'.format(batch_loss))


            if i and i % report_each == 0:
                batch_mean_loss = np.mean(epoch_loss[-report_each:])
                write_event(log, batch, batch_loss=batch_mean_loss, epoch=epoch, valid_loss=np.nan, val_iou=np.nan)
                save_images(outputs, labels, epoch, i)

                print('\n')
                print('Background mask: ', outputs[0, 0, :, :].min(), '\t', outputs[0, 0, :, :].max())
                print('Object mask: ', outputs[0, 1, :, :].min(), '\t', outputs[0, 1, :, :].max())

                #out_mask = torch.argmax(outputs, dim=1).float()
                #plt.imshow(out_mask[0, :, :])
                #plt.title("Out mask")
                #plt.show()

                #visualize_tensor(inputs[0, :, :, :], depths[0, :, :], out_mask[0, :, :], True)



            batch += 1
        # One epoch is finished, validation of epoch is up next.


        tq.close()

        ########## VALIDATION ##########
        model.eval()
        print('[VALIDATION]: Validation started \n')

        val_loss, iou_score = binary_validation(model, criterion, valid_loader, device)
        write_event(log, batch, batch_loss=np.nan, epoch=epoch, valid_loss=val_loss, val_iou=iou_score)
        valid_losses.append(val_loss)
        print('[VALIDATION]: Validation finished \n')

        epoch_mean = np.mean(epoch_loss)
        print(['[EPOCH FINISHED]\n'])

        save(epoch)

        print('{} Loss: {:.4f} IoU: {:.4f} Jaccard: {:.4f}'.format('epoch', epoch_mean, iou_score, jaccard))

        if epoch_mean < best_loss:
            best_loss = epoch_mean
            best_model_wts = copy.deepcopy(model.state_dict())

        train_loss_history.append(epoch_mean)

    # load best model weights
    best_model = model.load_state_dict(best_model_wts)

    return model, train_loss_history, best_model_wts, best_model
