import numpy as np
import torch
import matplotlib.pyplot as plt
import cv2
import torchvision
import random
import colorsys
from matplotlib import patches
from matplotlib.patches import Polygon
from skimage.measure import find_contours


# Tensor visualization
def denormalize(img):
    """
    Denormalizes the image
    :param img: input image tensor
    :return: denormalized image tensor
    """

    img[0] = img[0]*0.1872 + 0.2092
    img[1] = img[1]*0.1733 + 0.2172
    img[2] = img[2]*0.1625 + 0.2155
    return img


def mask_overlay(image, mask, color=(0, 255, 0)):
    """
    Helper function to visualize mask on the top of corresponding image
    :param image:
    :param mask:
    :param color:
    :return:
    """
    mask = np.dstack((mask, mask, mask)) * np.array(color)

    #mask = np.uint8(mask*255)
    image = np.uint8(image*255)

    #If binary segmentation:
    mask = np.uint8(mask)

    weighted_sum = cv2.addWeighted(mask, 0.5, image, 0.5, 0.)
    img = image.copy()
    ind = mask[:, :, 1] > 1
    img[ind] = weighted_sum[ind]
    return img


def visualize_mask(mask, title=None):
    mask = mask.detach().numpy().squeeze()
    plt.imshow(mask)
    plt.title(title)
    plt.show()


def apply_mask(image, mask, color, alpha=0.4):
    """Apply the given mask to the image.
    """
    for c in range(3):
        image[:, :, c] = np.where(mask == 1, image[:, :, c] * (1 - alpha) + alpha * color[c]*255, image[:, :, c])
    return image


def random_colors(N, bright=True):
    """
    Generate random colors.
    To get visually distinct colors, generate them in HSV space then
    convert to RGB.
    """
    brightness = 1.0 if bright else 0.7
    hsv = [(i / float(N), 1, brightness) for i in range(N)]
    colors = list(map(lambda c: colorsys.hsv_to_rgb(*c), hsv))
    random.shuffle(colors)
    return colors


def save_images(tensor, target, epoch, i):
    """
    Saves images during the training phase.
    :param tensor: input tensor of size [batch_size,classes,height,width]
    :param target: target tensor of shape [batch_size,height,width]
    :return: None
    """

    pred_mask = torch.argmax(tensor, dim=1)
    #torchvision.utils.save_image(tensor[0, 0, :, :].detach(), './plots/bg_mask' + str(epoch) + '-' + str(i) + '.png')
    torchvision.utils.save_image(target[0, :, :, :].squeeze(), './plots/gt_mask' + str(epoch) + '-' + str(i) + '.png')
    torchvision.utils.save_image(pred_mask[0, :, :], './plots/pred_mask' + str(epoch) + '-' + str(i) + '.png')


# Ndarray visualization

def display_instances(image, boxes, masks, class_ids, class_names,
                      scores=None, title="",
                      figsize=(10, 10), ax=None,
                      show_mask=True, show_bbox=True, show_class=True,
                      colors=None):
    """
    boxes: [num_instance, (y1, x1, y2, x2, class_id)] in image coordinates.
    masks: [height, width, num_instances]
    class_ids: [num_instances]
    class_names: list of class names of the dataset
    scores: (optional) confidence scores for each box
    title: (optional) Figure title
    show_mask, show_class, show_bbox: To show masks, classes and bounding boxes or not
    figsize: (optional) the size of the image
    colors: (optional) An array or colors to use with each object
    captions: (optional) A list of strings to use as captions for each object
    """
    # Number of instances
    N = boxes.shape[0]
    if not N:
        print("\n*** No instances to display *** \n")
    else:
        assert boxes.shape[0] == masks.shape[-1] == class_ids.shape[0]

    # If no axis is passed, create one and automatically call show()
    auto_show = False
    if not ax:
        fig, ax = plt.subplots(1, figsize=figsize)
        auto_show = True

    # Generate random colors
    colors = colors or random_colors(N)

    # Show area outside image boundaries.
    height, width = image.shape[:2]
    ax.axis('on')
    ax.set_title(title)

    masked_image = image.astype(np.uint32).copy()
    for i in range(N):
        color = colors[i]

        # Bounding box
        if not np.any(boxes[i]):
            # Skip this instance. Has no bbox. Likely lost in image cropping.
            continue
        y1, x1, y2, x2 = boxes[i]
        if show_bbox:
            p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2,
                                alpha=0.7, linestyle="dashed",
                                edgecolor=color, facecolor='none')
            ax.add_patch(p)

        # Label
        if not captions:
            class_id = class_ids[i]
            score = scores[i] if scores is not None else None
            label = class_names[class_id] if show_class else ""
            caption = "{} {:.3f}".format(label, score) if score else label
        else:
            caption = captions[i]
        ax.text(x1, y1 + 8, caption,
                color='w', size=11, backgroundcolor="none")

        # Mask
        mask = masks[:, :, i]
        if show_mask:
            masked_image = apply_mask(masked_image, mask, color)

        # Mask Polygon
        # Pad to ensure proper polygons for masks that touch image edges.
        padded_mask = np.zeros(
            (mask.shape[0] + 2, mask.shape[1] + 2), dtype=np.uint8)
        padded_mask[1:-1, 1:-1] = mask
        contours = find_contours(padded_mask, 0.5)
        for verts in contours:
            # Subtract the padding and flip (y, x) to (x, y)
            verts = np.fliplr(verts) - 1
            p = Polygon(verts, facecolor="none", edgecolor=color)
            ax.add_patch(p)

    ax.imshow(masked_image.astype(np.uint8))

    if auto_show:
        plt.title('Network output')
        plt.xlabel('pixels')
        plt.ylabel('pixels')
        plt.show()


#POINT CLOUD VISUALIZATION

def mask_pc_overlay(image, mask, color=(0, 255, 0)):
    """
    Helper function to visualize mask on the top of corresponding image
    :param image:
    :param mask:
    :param color:
    :return:
    """
    mask = mask * np.array(color)
    mask = np.uint8(mask*255)
    image = np.asarray(image)

    weighted_sum = cv2.addWeighted(mask, 0.5, image, 0.5, 0.)
    img = image.copy()
    ind = mask[:, :, 1] > 0
    img[ind] = weighted_sum[ind]
    img = Image(img.astype(np.uint8))
    return img


def visualize_pcd(pcd):
    """
    Flip and visualise point cloud
    :param pcd:
    :return:
    """
    pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
    draw_geometries([pcd])
